import logging
from typing import List, Optional, Set, TypeVar

import click

from tango.common.logging import click_logger
from tango.common.util import import_extra_module
from tango.server.workspace_server import WorkspaceServer
from tango.step_graph import StepGraph
from tango.workspace import Workspace

logger = logging.getLogger(__name__)


T = TypeVar("T")


class Executor:
    """
    An ``Executor`` is a class that is responsible for running steps and caching their results.
    """

    def __init__(
        self,
        workspace: Workspace,
        include_package: Optional[List[str]] = None,
        server: Optional[WorkspaceServer] = None,
    ) -> None:
        self.workspace = workspace
        self.include_package = include_package
        self.server = server

    def execute_step_graph(self, step_graph: StepGraph) -> str:
        """
        Execute a :class:`tango.step_graph.StepGraph`.
        """
        from tango import Step

        # Import included packages to find registered components.
        if self.include_package is not None:
            for package_name in self.include_package:
                import_extra_module(package_name)

        run_name = self.workspace.register_run(
            step for step in step_graph.values() if step.cache_results
        )
        if self.server is not None:
            click_logger.info(
                "Server started at "
                + click.style(self.server.address_for_display(run_name), bold=True)
            )

        ordered_steps = sorted(step_graph.values(), key=lambda step: step.name)

        # find uncacheable leaf steps
        interior_steps: Set[Step] = set()
        for step in ordered_steps:
            for dependency in step.dependencies:
                interior_steps.add(dependency)
        uncacheable_leaf_steps = {
            step for step in set(step_graph.values()) - interior_steps if not step.cache_results
        }

        click_logger.info(
            click.style("Starting new run ", fg="green")
            + click.style(run_name, fg="green", bold=True)
        )
        for step in ordered_steps:
            if step.cache_results:
                step.ensure_result(self.workspace)
            elif step in uncacheable_leaf_steps:
                step.result(self.workspace)

        # Print everything that has been computed.
        for step in ordered_steps:
            if step in self.workspace.step_cache:
                info = self.workspace.step_info(step)
                click_logger.info(
                    click.style("\N{check mark} The output for ", fg="green")
                    + click.style(f'"{step.name}"', bold=True, fg="green")
                    + click.style(" is in ", fg="green")
                    + click.style(f"{info.result_location}", bold=True, fg="green")
                )

        return run_name



from dataclasses import dataclass, field
from typing import Any, Generic, Iterable, Iterator, Mapping, Sequence, TypeVar

T = TypeVar("T")
S = TypeVar("S")


@dataclass
class DatasetDictBase(Generic[S], Mapping[str, S]):
    """
    The base class for :class:`DatasetDict` and :class:`IterableDatasetDict`.
    """

    splits: Mapping[str, S]
    """
    A mapping of dataset split names to splits.
    """

    metadata: Mapping[str, Any] = field(default_factory=dict)
    """
    Metadata can contain anything you need.
    """

    def __getitem__(self, split: str) -> S:
        """
        Get a split in :attr:`splits`.
        """
        return self.splits[split]

    def __contains__(self, split: str) -> bool:  # type: ignore[override]
        """
        Checks if :attr:`splits` contains the given split.
        """
        return split in self.splits

    def __iter__(self) -> Iterator[str]:
        """
        Returns an iterator over the keys in :attr:`splits`.
        """
        return iter(self.splits.keys())

    def __len__(self) -> int:
        """
        Returns the number of splits in :attr:`splits`.
        """
        return len(self.splits)

    def keys(self):
        """
        Returns the split names in :attr:`splits`.
        """
        return self.splits.keys()


@dataclass
class DatasetDict(DatasetDictBase[Sequence[T]], Generic[T]):
    """
    A generic :class:`~collections.abc.Mapping` class of split names (:class:`str`) to datasets
    (``Sequence[T]``).
    """


@dataclass
class IterableDatasetDict(DatasetDictBase[Iterable[T]], Generic[T]):
    """
    An "iterable" version of :class:`DatasetDict`, where the dataset splits have
    type ``Iterable[T]`` instead of ``Sequence[T]``. This is useful for streaming datasets.
    """



"""
.. important::
    To use this integration you should install ``tango`` with the "pytorch_lightning" extra
    (e.g. ``pip install tango[pytorch_lightning]``) or just install PyTorch Lightning after the fact.

Components for Tango integration with `PyTorch Lightning <https://www.pytorchlightning.ai/>`_.

These include a basic training loop :class:`~tango.step.Step` and registrable versions
of ``pytorch_lightning`` classes, such as :class:`~pytorch_lightning.core.lightning.LightningModule`,
:class:`~pytorch_lightning.callbacks.base.Callback`, etc.

Example: training a model
-------------------------

Let's look a simple example of training a model.

We'll make a very basic regression model and generate some fake data to train on.
First, the setup:

.. testcode::

    import torch
    import torch.nn as nn

    from tango.common.dataset_dict import DatasetDict
    from tango.step import Step
    from tango.integrations.pytorch_lightning import LightningModule

Now let's build and register our model:

.. testcode::

    @LightningModule.register("basic_regression")
    class BasicRegression(LightningModule):
        def __init__(self):
            super().__init__()
            self.linear = nn.Linear(10, 1)
            self.sigmoid = nn.Sigmoid()
            self.mse = nn.MSELoss()

        def forward(self, x, y=None):
            pred = self.sigmoid(self.linear(x))
            out = {"pred": pred}
            if y is not None:
                out["loss"] = self.mse(pred, y)
            return out

        def _to_params(self):
            return {}

        def training_step(self, batch, batch_idx):
            outputs = self.forward(**batch)
            return outputs["loss"]

        def validation_step(self, batch, batch_idx, dataloader_idx=0):
            outputs = self.forward(**batch)
            return outputs

        def configure_optimizers(self):
            optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
            return optimizer

Lastly, we'll need a step to generate data:

.. testcode::

    @Step.register("generate_data")
    class GenerateData(Step):
        DETERMINISTIC = True
        CACHEABLE = False

        def run(self) -> DatasetDict:
            torch.manual_seed(1)
            return DatasetDict(
                {
                    "train": [{"x": torch.rand(10), "y": torch.rand(1)} for _ in range(64)],
                    "validation": [{"x": torch.rand(10), "y": torch.rand(1)} for _ in range(32)],
                }
            )

You could then run this experiment with a config that looks like this:

.. literalinclude:: ../../../../test_fixtures/integrations/pytorch_lightning/train.jsonnet

.. testcode::
    :hide:

    from tango.common.testing import run_experiment
    from tango.common.registrable import Registrable

    with run_experiment(
        "test_fixtures/integrations/pytorch_lightning/train.jsonnet"
    ) as run_dir:
        assert (run_dir / "train").is_dir(), "Output for the 'train' step was not produced."
    # Restore state of registry.
    del Registrable._registry[Step]["generate_data"]
    del Registrable._registry[LightningModule]["basic_regression"]

For example,

.. code-block::

    tango run train.jsonnet -i my_package -d /tmp/train

.. testoutput::
    :options: +ELLIPSIS

    Starting new run ...
    ● Starting step "data" (needed by "train") ...
    ✓ Finished step "data"
    ● Starting step "train" ...
    ...
    ✓ Finished step "train"
    ✓ The output for "train" is in ...

Tips
----

PyTorch Lightning functionality
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

You can use existing Pytorch Lightning callbacks, loggers, and profilers, which are registered as
"pytorch_lightning::name of callback logger or profiler".
For example, the :class:`~pytorch_lightning.callbacks.early_stopping.EarlyStopping` callback
is registered as "pytorch_lightning::EarlyStopping".

"""

__all__ = [
    "LightningAccelerator",
    "LightningCallback",
    "LightningDataModule",
    "LightningLogger",
    "LightningModule",
    "LightningProfiler",
    "LightningTrainingTypePlugin",
    "LightningPrecisionPlugin",
    "LightningClusterEnvironmentPlugin",
    "LightningCheckpointPlugin",
    "LightningTrainStep",
    "LightningTrainer",
]

from .accelerators import LightningAccelerator
from .callbacks import LightningCallback
from .data import LightningDataModule
from .loggers import LightningLogger
from .model import LightningModule
from .plugins import (
    LightningCheckpointPlugin,
    LightningClusterEnvironmentPlugin,
    LightningPrecisionPlugin,
    LightningTrainingTypePlugin,
)
from .profilers import LightningProfiler
from .train import LightningTrainer, LightningTrainStep



import logging
import os
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

from tango.common.aliases import PathOrStr
from tango.common.logging import initialize_logging


class TangoTestCase:
    """
    A custom testing class that

    * disables some of the more verbose logging,
    * creates and destroys a temp directory as a test fixture, and
    * restores the internal state of the `Registrable` class at the end of each test method.

    """

    PROJECT_ROOT = (Path(__file__).parent / ".." / "..").resolve()
    """
    Root of the git repository.
    """

    MODULE_ROOT = PROJECT_ROOT / "tango"
    """
    Root of the tango module.
    """

    TESTS_ROOT = PROJECT_ROOT / "tests"
    """
    Root of the tests directory.
    """

    FIXTURES_ROOT = PROJECT_ROOT / "test_fixtures"
    """
    Root of the test fixtures directory.
    """

    def setup_method(self):
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s", level=logging.DEBUG
        )

        # Disabling some of the more verbose logging statements that typically aren't very helpful
        # in tests.
        logging.getLogger("urllib3.connectionpool").disabled = True

        # Create a temporary scratch directory.
        self.TEST_DIR = Path(tempfile.mkdtemp(prefix="tango_tests"))
        os.makedirs(self.TEST_DIR, exist_ok=True)

    def teardown_method(self):
        shutil.rmtree(self.TEST_DIR)

    def run(
        self,
        config: Union[PathOrStr, Dict[str, Any]],
        overrides: Optional[Union[Dict[str, Any], str]] = None,
        include_package: Optional[List[str]] = None,
    ) -> Path:
        from tango.__main__ import TangoGlobalSettings, _run

        from .params import Params

        if isinstance(config, dict):
            params = Params(config)
            config = self.TEST_DIR / "config.json"
            params.to_file(cast(Path, config))

        if isinstance(overrides, dict):
            import json

            overrides = json.dumps(overrides)

        return _run(
            TangoGlobalSettings(),
            str(config),
            workspace_dir=str(self.TEST_DIR / "workspace"),
            overrides=overrides,
            include_package=include_package,
            start_server=False,
        )


@contextmanager
def run_experiment(
    config: Union[PathOrStr, Dict[str, Any]], overrides: Optional[Union[Dict[str, Any], str]] = None
):
    """
    A context manager to make testing experiments easier. On ``__enter__`` it runs
    the experiment and returns the path to the cache directory, a temporary directory that will be
    cleaned up on ``__exit__``.
    """
    initialize_logging(enable_click_logs=True)
    test_case = TangoTestCase()
    try:
        test_case.setup_method()
        yield test_case.run(config, overrides=overrides)
    finally:
        test_case.teardown_method()



import torch

from tango.common.registrable import Registrable


class Model(torch.nn.Module, Registrable):
    """
    This is simply a :class:`~tango.common.Registrable` :class:`torch.nn.Module`
    that returns a :class:`dict` from its :meth:`~torch.nn.Module.forward()` method which
    should include the ``loss`` during training and validation.
    """



from typing import Union

import pytorch_lightning as pl
from pytorch_lightning.plugins.environments.cluster_environment import (
    ClusterEnvironment,
)
from pytorch_lightning.plugins.io.checkpoint_plugin import CheckpointIO
from pytorch_lightning.plugins.precision.precision_plugin import PrecisionPlugin
from pytorch_lightning.plugins.training_type.training_type_plugin import (
    TrainingTypePlugin,
)

from tango.common.registrable import Registrable


class LightningTrainingTypePlugin(TrainingTypePlugin, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning
    :class:`~pytorch_lightning.plugins.training_type.training_type_plugin.TrainingTypePlugin`
    class.
    """


class LightningPrecisionPlugin(PrecisionPlugin, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning
    :class:`~pytorch_lightning.plugins.precision.precision_plugin.PrecisionPlugin`
    class.
    """


class LightningClusterEnvironmentPlugin(ClusterEnvironment, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning
    :class:`~pytorch_lightning.plugins.environments.cluster_environment.ClusterEnvironment`
    class.
    """


class LightningCheckpointPlugin(CheckpointIO, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning
    :class:`~pytorch_lightning.plugins.o.checkpoint_plugin.CheckpointIO`
    class.
    """


# Register all plugins.
for name, cls in pl.plugins.training_type.__dict__.items():
    if (
        isinstance(cls, type)
        and issubclass(cls, TrainingTypePlugin)
        and not cls == TrainingTypePlugin
    ):
        LightningTrainingTypePlugin.register("pytorch_lightning::" + name)(cls)

for name, cls in pl.plugins.precision.__dict__.items():
    if isinstance(cls, type) and issubclass(cls, PrecisionPlugin) and not cls == PrecisionPlugin:
        LightningPrecisionPlugin.register("pytorch_lightning::" + name)(cls)

for name, cls in pl.plugins.environments.__dict__.items():
    if (
        isinstance(cls, type)
        and issubclass(cls, ClusterEnvironment)
        and not cls == ClusterEnvironment
    ):
        LightningClusterEnvironmentPlugin.register("pytorch_lightning::" + name)(cls)

for name, cls in pl.plugins.io.__dict__.items():
    if isinstance(cls, type) and issubclass(cls, CheckpointIO) and not cls == CheckpointIO:
        LightningCheckpointPlugin.register("pytorch_lightning::" + name)(cls)


ALL_PLUGIN_TYPES = Union[
    LightningTrainingTypePlugin,
    LightningPrecisionPlugin,
    LightningClusterEnvironmentPlugin,
    LightningCheckpointPlugin,
]



from typing import Any

from tango.step import Step


@Step.register("print")
class PrintStep(Step):
    """
    This step just prints out its input and also returns what it prints.
    """

    DETERMINISTIC = True
    CACHEABLE = False  # so fast it's not worth caching

    def run(self, input: Any) -> str:  # type: ignore[override]
        """
        Print out the input.
        """
        out = str(input)
        print(out)
        return out



import os
import shutil
from os import PathLike
from typing import Any, Iterable, MutableSequence, Union

from sqlitedict import SqliteDict

from tango.common.sequences import SlicedSequence


class SqliteSparseSequence(MutableSequence[Any]):
    def __init__(self, filename: Union[str, PathLike], read_only: bool = False):
        self.table = SqliteDict(filename, "sparse_sequence", flag="r" if read_only else "c")

    def __del__(self):
        self.close()

    def __getitem__(self, i: Union[int, slice]) -> Any:
        if isinstance(i, int):
            try:
                return self.table[str(i)]
            except KeyError:
                current_length = len(self)
                if i >= current_length or current_length <= 0:
                    raise IndexError("list index out of range")
                elif i < 0 < current_length:
                    return self.__getitem__(i % current_length)
                else:
                    return None
        elif isinstance(i, slice):
            return SlicedSequence(self, i)
        else:
            raise TypeError(f"list indices must be integers or slices, not {i.__class__.__name__}")

    def __setitem__(self, i: Union[int, slice], value: Any):
        if isinstance(i, int):
            current_length = len(self)
            if i < 0:
                i %= current_length
            self.table[str(i)] = value
            self.table["_len"] = max(i, current_length)
            self.table.commit()
        else:
            raise TypeError(f"list indices must be integers, not {i.__class__.__name__}")

    def __delitem__(self, i: Union[int, slice]):
        current_length = len(self)
        if isinstance(i, int):
            if i < 0:
                i %= current_length
            if i >= current_length:
                raise IndexError("list assignment index out of range")
            for index in range(i + 1, current_length):
                self.table[str(index - 1)] = self.table.get(str(index))
            del self.table[str(current_length - 1)]
            self.table["_len"] = current_length - 1
            self.table.commit()
        elif isinstance(i, slice):
            # This isn't very efficient for continuous slices.
            for index in reversed(range(*i.indices(current_length))):
                del self[index]
        else:
            raise TypeError(f"list indices must be integers or slices, not {i.__class__.__name__}")

    def extend(self, values: Iterable[Any]) -> None:
        current_length = len(self)
        index = -1
        for index, value in enumerate(values):
            self.table[str(index + current_length)] = value
        if index < 0:
            return
        self.table["_len"] = current_length + index + 1
        self.table.commit()

    def insert(self, i: int, value: Any) -> None:
        current_length = len(self)
        for index in reversed(range(i, current_length)):
            self.table[str(index + 1)] = self.table.get(str(index))
        self.table[str(i)] = value
        self.table["_len"] = current_length + 1
        self.table.commit()

    def __len__(self) -> int:
        try:
            return self.table["_len"]
        except KeyError:
            return 0

    def clear(self) -> None:
        self.table.clear()
        self.table.commit()

    def close(self) -> None:
        if self.table is not None:
            self.table.close()
            self.table = None

    def copy_to(self, target: Union[str, PathLike]):
        try:
            os.link(self.table.filename, target)
        except OSError as e:
            if e.errno == 18:  # Cross-device link
                shutil.copy(self.table.filename, target)
            else:
                raise



import os
import sys
from typing import Any, Dict, List, Optional

from tango.integrations.torch.train_callback import TrainCallback


@TrainCallback.register("wandb::log")
class WandbTrainCallback(TrainCallback):
    """
    A torch :class:`~tango.integrations.torch.TrainCallback` for use with
    the :class:`~tango.integrations.torch.TorchTrainStep` that logs training and
    validation metrics to W&B.

    .. tip::

        Registered as a :class:`~tango.integrations.torch.TrainCallback`
        under the name "wandb::log".

    .. important::
        This class also requires the :mod:`~tango.integrations.torch` integration.

    Parameters
    ----------

    project : :class:`str`
        W&B project to associated this run with.
    entity : :class:`str`
        W&B entity (user or organization) to associated this run with.
    group : :class:`str`
        W&B group to associated this run with.
    name : :class:`str`
        Set the name of the run in W&B.
    notes : :class:`str`
        Arbitrary notes to add in W&B to this run.
    tags : ``List[str]``
        Arbitrary tags to add in W&B to this run.
    watch_model : :class:`bool`
        If ``True``, ``wandb.watch()`` is called to collect gradients and other information
        about the model throughout training.
        See `docs.wandb.ai/ref/python/watch <https://docs.wandb.ai/ref/python/watch>`_.
    wandb_config : ``Dict[str, Any]``
        Arbitrary configuration fields to set in W&B for this run.
        See `docs.wandb.ai/guides/track/config <https://docs.wandb.ai/guides/track/config>`_.
    wandb_kwargs : ``Dict[str, Any]``
        Additional keyword arguments that will be passed to ``wandb.init()``.
        See `docs.wandb.ai/ref/python/init <https://docs.wandb.ai/ref/python/init>`_.

    """

    def __init__(
        self,
        *args,
        project: Optional[str] = None,
        entity: Optional[str] = None,
        group: Optional[str] = None,
        name: Optional[str] = None,
        notes: Optional[str] = None,
        tags: Optional[List[str]] = None,
        watch_model: bool = False,
        wandb_config: Optional[Dict[str, Any]] = None,
        wandb_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        if self.is_local_main_process and "WANDB_API_KEY" not in os.environ:
            print(
                "Missing environment variable 'WANDB_API_KEY' required to authenticate to Weights & Biases.",
                file=sys.stderr,
            )

        _wandb_config = self.train_config.as_dict()
        _wandb_config["work_dir"] = str(self.train_config.work_dir.resolve())
        if wandb_config is not None:
            _wandb_config.update(wandb_config)

        self._watch_model = watch_model
        self._run_id: Optional[str] = None
        self._wandb_kwargs: Dict[str, Any] = dict(
            dir=str(self.work_dir),
            project=project,
            entity=entity,
            group=group,
            name=name or self._get_run_name_from_step(),
            notes=notes,
            config=_wandb_config,
            tags=tags,
            anonymous="allow",
            **(wandb_kwargs or {}),
        )

    def _get_run_name_from_step(self) -> str:
        """
        Generate a run name for W&B based on the step unique ID.
        """
        step_name, step_hash = self.train_config.step.split("-", maxsplit=1)
        return f"{step_name}-{step_hash[:6]}"

    def state_dict(self) -> Dict[str, Any]:
        if self.is_local_main_process:
            return {
                "run_id": self._run_id,
            }
        else:
            return {}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        if self.is_local_main_process:
            self._wandb_kwargs["resume"] = "auto"
            self._run_id = state_dict["run_id"]

    def pre_train_loop(self) -> None:
        if self.is_local_main_process:
            import wandb

            self.wandb = wandb

            if self._run_id is None:
                self._run_id = self.wandb.util.generate_id()

            self.wandb.init(id=self._run_id, **self._wandb_kwargs)

            if self._watch_model:
                self.wandb.watch(self.model)

    def post_train_loop(self) -> None:
        if self.is_local_main_process:
            self.wandb.finish()

    def log_batch(self, step: int, batch_loss: float) -> None:
        if self.is_local_main_process:
            self.wandb.log(
                {"train/loss": batch_loss, "train/lr": self.optimizer.param_groups[0]["lr"]},
                step=step + 1,
            )

    def post_val_loop(self, step: int, val_metric: float, best_val_metric: float) -> None:
        if self.is_local_main_process:
            self.wandb.log(
                {
                    f"val/{self.train_config.val_metric_name}": val_metric,
                    f"val/best_{self.train_config.val_metric_name}": best_val_metric,
                },
                step=step + 1,
            )



import pytorch_lightning as pl

from tango.common.registrable import Registrable


class LightningCallback(pl.callbacks.base.Callback, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning :class:`~pytorch_lightning.callbacks.base.Callback`.
    """


# Register all callbacks.
for name, cls in pl.callbacks.__dict__.items():
    if (
        isinstance(cls, type)
        and issubclass(cls, pl.callbacks.base.Callback)
        and not cls == pl.callbacks.base.Callback
    ):
        LightningCallback.register("pytorch_lightning::" + name)(cls)



import importlib
import pkgutil
import signal
import string
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Optional, Set, Union

from .aliases import PathOrStr
from .exceptions import SigTermReceived


def _handle_sigterm(sig, frame):
    raise SigTermReceived


def install_sigterm_handler():
    signal.signal(signal.SIGTERM, _handle_sigterm)


@contextmanager
def push_python_path(path: PathOrStr):
    """
    Prepends the given path to `sys.path`.

    This method is intended to use with `with`, so after its usage, its value willbe removed from
    `sys.path`.
    """
    # In some environments, such as TC, it fails when sys.path contains a relative path, such as ".".
    path = Path(path).resolve()
    path = str(path)
    sys.path.insert(0, path)
    try:
        yield
    finally:
        # Better to remove by value, in case `sys.path` was manipulated in between.
        sys.path.remove(path)


_extra_imported_modules: Set[str] = set()


def get_extra_imported_modules() -> Set[str]:
    return _extra_imported_modules


def import_extra_module(package_name: str) -> None:
    global _extra_imported_modules
    import_module_and_submodules(package_name)
    _extra_imported_modules.add(package_name)


def import_module_and_submodules(package_name: str, exclude: Optional[Set[str]] = None) -> None:
    """
    Import all submodules under the given package.

    Primarily useful so that people using tango can specify their own custom packages
    and have their custom classes get loaded and registered.
    """
    if exclude and package_name in exclude:
        return

    importlib.invalidate_caches()

    # For some reason, python doesn't always add this by default to your path, but you pretty much
    # always want it when using `--include-package`.  And if it's already there, adding it again at
    # the end won't hurt anything.
    with push_python_path("."):
        # Import at top level
        module = importlib.import_module(package_name)
        path = getattr(module, "__path__", [])
        path_string = "" if not path else path[0]

        # walk_packages only finds immediate children, so need to recurse.
        for module_finder, name, _ in pkgutil.walk_packages(path):
            # Sometimes when you import third-party libraries that are on your path,
            # `pkgutil.walk_packages` returns those too, so we need to skip them.
            if path_string and module_finder.path != path_string:  # type: ignore[union-attr]
                continue
            subpackage = f"{package_name}.{name}"
            import_module_and_submodules(subpackage, exclude=exclude)


def _parse_bool(value: Union[bool, str]) -> bool:
    if isinstance(value, bool):
        return value
    if value in {"1", "true", "True", "TRUE"}:
        return True
    return False


def find_submodules(
    module: Optional[str] = None,
    match: Optional[Set[str]] = None,
    exclude: Optional[Set[str]] = None,
    recursive: bool = True,
) -> Iterable[str]:
    """
    Find tango submodules.
    """
    from fnmatch import fnmatch

    root = Path(__file__).parent / ".."
    if module:
        if module.startswith("tango."):
            module = module.replace("tango.", "", 1)
        for m in module.split("."):
            root = root / m
        module = f"tango.{module}"
    else:
        module = "tango"
    for path in root.iterdir():
        if path.name[0] in {"_", "."}:
            continue
        submodule: str
        if path.is_dir():
            submodule = path.name
        elif path.suffix == ".py":
            submodule = path.name[:-3]
        else:
            continue
        submodule = f"{module}.{submodule}"
        if exclude and any((fnmatch(submodule, pat) for pat in exclude)):
            continue
        if match and not any((fnmatch(submodule, pat) for pat in match)):
            continue
        yield submodule
        if recursive and path.is_dir():
            yield from find_submodules(submodule, match=match, exclude=exclude)


def find_integrations() -> Iterable[str]:
    """
    Find all tango integration modules.
    """
    yield from find_submodules("tango.integrations", recursive=False)


SAFE_FILENAME_CHARS = frozenset("-_.%s%s" % (string.ascii_letters, string.digits))


def filename_is_safe(filename: str) -> bool:
    return all(c in SAFE_FILENAME_CHARS for c in filename)


def could_be_class_name(name: str) -> bool:
    if "." in name and not name.endswith("."):
        return all([_is_valid_python_name(part) for part in name.split(".")])
    else:
        return False


def _is_valid_python_name(name: str) -> bool:
    return bool(name and name[0].isalpha() and name.isalnum())



from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

import torch

from tango.common.lazy import Lazy
from tango.common.registrable import Registrable

T = TypeVar("T")


class DataCollator(Generic[T], Registrable):
    """
    A :class:`~tango.common.Registrable` version of a ``collate_fn``
    for a ``DataLoader``.

    Subclasses just need to implement :meth:`__call__()`.
    """

    default_implementation = "concat_tensor_dicts"
    """
    The default implementation is :class:`ConcatTensorDictsCollator`.
    """

    def __call__(self, items: List[T]) -> Dict[str, Any]:
        """
        Takes a list of items from a dataset and combines them into a batch.
        """
        raise NotADirectoryError


@DataCollator.register("concat_tensor_dicts")
class ConcatTensorDictsCollator(DataCollator[Dict[str, Any]]):
    """
    A simple ``collate_fn`` that expects items to be dictionaries of tensors.
    The tensors are just concatenated together.

    .. tip::

        Registered as a :class:`DataCollator` under the name "concat_tensor_dicts".
    """

    def __call__(self, items: List[Dict[str, Any]]) -> Dict[str, Any]:
        out = {}
        keys = items[0].keys()
        for key in keys:
            if isinstance(items[0][key], torch.Tensor):
                out[key] = torch.cat([item[key].unsqueeze(0) for item in items])
            else:
                out[key] = [item[key] for item in items]  # type: ignore[assignment]
        return out


class Sampler(torch.utils.data.Sampler, Registrable):
    """
    A :class:`~tango.common.Registrable` version of a PyTorch
    :class:`~torch.utils.data.Sampler`.

    All `built-in PyTorch samplers
    <https://pytorch.org/docs/stable/data.html#data-loading-order-and-sampler>`_
    are registered under their corresponding class name (e.g. "RandomSampler").
    """


@Sampler.register("BatchSampler")
class BatchSampler(torch.utils.data.BatchSampler, Sampler):
    def __init__(self, sampler: Sampler, batch_size: int, drop_last: bool) -> None:
        super().__init__(sampler, batch_size, drop_last)


# Register all remaining samplers.
for name, cls in torch.utils.data.__dict__.items():
    if (
        isinstance(cls, type)
        and issubclass(cls, torch.utils.data.Sampler)
        and not cls == torch.utils.data.Sampler
        and name not in Sampler.list_available()
    ):
        Sampler.register("torch::" + name)(cls)


class DataLoader(torch.utils.data.DataLoader, Registrable):
    """
    A :class:`~tango.common.Registrable` version of a PyTorch
    :class:`~torch.utils.data.DataLoader`.
    """

    default_implementation = "default"

    def __init__(
        self,
        dataset: torch.utils.data.Dataset,
        collate_fn: Optional[DataCollator] = ConcatTensorDictsCollator(),
        sampler: Optional[Union[Lazy[Sampler], Sampler]] = None,
        **kwargs
    ):
        super().__init__(
            dataset,
            collate_fn=collate_fn,
            sampler=sampler.construct(data_source=dataset, dataset=dataset)
            if isinstance(sampler, Lazy)
            else sampler,
            **kwargs
        )


DataLoader.register("default")(DataLoader)



import pytorch_lightning as pl

from tango.common.registrable import Registrable


class LightningDataModule(pl.LightningDataModule, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    :class:`pytorch_lightning.LightningDataModule`. It includes the following
    methods:

    * :meth:`~pytorch_lightning.extensions.datamodules.LightningDataModule.prepare_data()`
    * :meth:`~pytorch_lightning.extensions.datamodules.LightningDataModule.setup()`
    * :meth:`~pytorch_lightning.extensions.datamodules.LightningDataModule.train_dataloader()`
    * :meth:`~pytorch_lightning.extensions.datamodules.LightningDataModule.val_dataloader()`
    * :meth:`~pytorch_lightning.extensions.datamodules.LightningDataModule.test_dataloader()`
    """

    def _to_params(self):
        return {}



from pathlib import Path
from typing import Any, Dict, List, Optional

from tango.common.dataset_dict import DatasetDictBase
from tango.common.registrable import Registrable

from .data import DataLoader
from .exceptions import StopEarly
from .model import Model
from .optim import LRScheduler, Optimizer
from .train_config import TrainConfig


class TrainCallback(Registrable):
    """
    A ``TrainCallback`` is a :class:`~tango.common.Registrable` class
    that can be used within :class:`TorchTrainStep` to customize behavior in the training
    loop.

    .. tip::
        All of the parameters to this base class will be automatically set within
        the training loop, so you shouldn't include them in your config for your callbacks.

    .. seealso::
        See :class:`~tango.integrations.wandb.WandbTrainCallback` for an example
        implementation.

    Attributes
    ----------
    train_config : :class:`TrainConfig`
    model : :class:`Model`
    optimizer : :class:`Optimizer`
    dataset_dict : :class:`tango.common.DatasetDictBase`
    train_dataloader : :class:`DataLoader`
    validation_dataloader : :class:`DataLoader`, optional
    lr_scheduler : :class:`LRScheduler`, optional

    """

    def __init__(
        self,
        train_config: TrainConfig,
        model: Model,
        optimizer: Optimizer,
        dataset_dict: DatasetDictBase,
        train_dataloader: DataLoader,
        validation_dataloader: Optional[DataLoader] = None,
        lr_scheduler: Optional[LRScheduler] = None,
    ) -> None:
        self.train_config = train_config
        self.model = model
        self.optimizer = optimizer
        self.dataset_dict = dataset_dict
        self.train_dataloader = train_dataloader
        self.validation_dataloader = validation_dataloader
        self.lr_scheduler = lr_scheduler

    @property
    def work_dir(self) -> Path:
        """
        The working directory of the current train step.
        """
        return self.train_config.work_dir

    @property
    def is_local_main_process(self) -> bool:
        """
        If the current worker is the main distributed worker of the current node.
        """
        return self.train_config.is_local_main_process

    def state_dict(self) -> Dict[str, Any]:
        """
        Return any state that needs to be kept after a restart.
        """
        return {}

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        """
        Load the state on a restart.
        """
        pass

    def pre_train_loop(self) -> None:
        """
        Called right before the first batch is processed, or after a restart.
        """
        pass

    def post_train_loop(self) -> None:
        """
        Called after the training loop completes.

        This is the last method that is called, so any cleanup can be done in this method.
        """
        pass

    def pre_epoch(self, epoch: int) -> None:
        """
        Called right before the start of an epoch. Epochs start at 0.
        """
        pass

    def post_epoch(self, epoch: int) -> None:
        """
        Called after an epoch is completed. Epochs start at 0.
        """
        pass

    def pre_checkpoint(self, checkpoint_state: Dict[str, Any]) -> None:
        """
        Called directly before the checkpoint is saved.
        """
        pass

    def post_checkpoint(self, checkpoint_path: Path) -> None:
        """
        Called directly after a checkpoint is saved.
        """
        pass

    def pre_batch(self, step: int, batch: List[Dict[str, Any]]) -> None:
        """
        Called directly before processing a batch.

        .. note::
            A type of ``batch`` is a list because with gradient accumulation there will
            more than one "micro batch" in the batch.

        """
        pass

    def post_batch(self, step: int, batch_loss: float) -> None:
        """
        Called directly after processing a batch, but before unscaling gradients,
        clipping gradients, and taking an optimizer step.

        .. note::
            The ``batch_loss`` here is the loss local to the current worker, not the
            overall (average) batch loss across distributed workers.

            If you need the average loss, use :meth:`log_batch()`.

        """
        pass

    def log_batch(self, step: int, batch_loss: float) -> None:
        """
        Called after the optimizer step. Here ``batch_loss`` is the average loss across
        all distributed workers.

        .. note::
            This callback method is not necessarily called on every step.
            The frequency depends on the value of the ``log_every`` parameter of
            :class:`TorchTrainStep`.

        """
        pass

    def pre_val_batch(self, step: int, val_step: int, val_batch: Dict[str, Any]) -> None:
        """
        Called right before a validation batch is processed.
        """
        pass

    def post_val_batch(self, step: int, val_step: int, val_batch_outputs: Dict[str, Any]) -> None:
        """
        Called right after a validation batch is processed with the outputs of the batch.
        """
        pass

    def post_val_loop(self, step: int, val_metric: float, best_val_metric: float) -> None:
        """
        Called right after the validation loop finishes.
        """
        pass


@TrainCallback.register("torch::stop_early")
class StopEarlyCallback(TrainCallback):
    """
    A :class:`TrainCallback` for early stopping. Training is stopped early after
    ``patience`` steps without an improvement to the validation metric.

    .. tip::

        Registered as a :class:`TrainCallback` under the name "torch::stop_early".
    """

    def __init__(self, *args, patience: int = 10000, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.patience = patience
        self.best_step = 0

    def post_val_loop(self, step: int, val_metric: float, best_val_metric: float) -> None:
        if val_metric == best_val_metric:
            self.best_step = step
        elif step > self.best_step + self.patience:
            raise StopEarly



from collections import defaultdict
from pathlib import Path

from setuptools import find_packages, setup


def parse_requirements_file(path, allowed_extras: set = None, include_all_extra: bool = True):
    requirements = []
    extras = defaultdict(list)
    with open(path) as requirements_file:
        import re

        def fix_url_dependencies(req: str) -> str:
            """Pip and setuptools disagree about how URL dependencies should be handled."""
            m = re.match(
                r"^(git\+)?(https|ssh)://(git@)?github\.com/([\w-]+)/(?P<name>[\w-]+)\.git", req
            )
            if m is None:
                return req
            else:
                return f"{m.group('name')} @ {req}"

        for line in requirements_file:
            line = line.strip()
            if line.startswith("#") or len(line) <= 0:
                continue
            req, *needed_by = line.split("# needed by:")
            req = fix_url_dependencies(req.strip())
            if needed_by:
                for extra in needed_by[0].strip().split(","):
                    extra = extra.strip()
                    if allowed_extras is not None and extra not in allowed_extras:
                        raise ValueError(f"invalid extra '{extra}' in {path}")
                    extras[extra].append(req)
                if include_all_extra and req not in extras["all"]:
                    extras["all"].append(req)
            else:
                requirements.append(req)
    return requirements, extras


# Find all integrations.
integrations = set(
    p.name
    for p in Path("tango/integrations").glob("*")
    if p.is_dir() and not p.name[0] in {".", "_"}
)

# Load requirements.
install_requirements, extras = parse_requirements_file(
    "requirements.txt", allowed_extras=integrations
)
dev_requirements, dev_extras = parse_requirements_file(
    "dev-requirements.txt", allowed_extras={"examples"}, include_all_extra=False
)
extras["dev"] = dev_requirements
extras.update(dev_extras)

# Validate extras.
assert "all" in extras
assert "dev" in extras
assert "examples" in extras
for integration in integrations:
    assert integration in extras

# version.py defines the VERSION and VERSION_SHORT variables.
# We use exec here so we don't import `cached_path` whilst setting up.
VERSION = {}  # type: ignore
with open("tango/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

setup(
    name="ai2-tango",
    version=VERSION["VERSION"],
    description="A library for choreographing your machine learning research.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="",
    url="https://github.com/allenai/tango",
    author="Allen Institute for Artificial Intelligence",
    author_email="contact@allenai.org",
    license="Apache",
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "test_fixtures", "test_fixtures.*"],
    ),
    package_data={"tango.server": ["*.svg", "*.css", "*.js", "*.html"]},
    entry_points={"console_scripts": ["tango=tango.__main__:main"]},
    install_requires=install_requirements,
    extras_require=extras,
    python_requires=">=3.7.1",
)



import pytorch_lightning as pl

from tango.common.registrable import Registrable


class LightningAccelerator(pl.accelerators.Accelerator, Registrable):
    """
    This is simply a :class:`~tango.common.registrable.Registrable`
    version of the PyTorch Lightning :class:`~pytorch_lightning.accelerators.Accelerator`.
    """


# Register all accelerators.
for name, cls in pl.accelerators.__dict__.items():
    if (
        isinstance(cls, type)
        and issubclass(cls, pl.accelerators.Accelerator)
        and not cls == pl.accelerators.Accelerator
    ):
        LightningAccelerator.register("pytorch_lightning::" + name)(cls)
